/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This file is the NBTV generator*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "nbtv_format.h"
#include "gui.h"
#include <math.h>
#include "input.h"
#include "nbtv_gen.h"
#include "preview.h"
#include "output.h"

double target_timestamp; /*Time to run until*/
char nbtv_gen; /*Run flag*/
//char frame_complete;
int frame_count = 0;
//int target_frame_count;
char frame_changed, format_changed, repeat_frame;
SDL_Surface *window;
void *frame_samples;

uint8_t buffer[65535], key[1<<17];
float float_samples[4][1<<17];
char channel_zeroed[4];

int frame_len;
/*Gets a pixel. This also does orientation transformation and color stuff*/
double get_converted_pixel(SDL_Surface *input, struct _input *p, struct NBTV_format *format, int channel, int line, int n){
	int offset;
	double R,G,B;
	double ov;
	uint8_t *pixels = input->pixels;
	R = log10((abs(format->channels[channel].red_gain)+1)*10);
	if(format->channels[channel].red_gain < 0){R=1/R;}
	G = log10((abs(format->channels[channel].green_gain)+1)*10);
	if(format->channels[channel].green_gain < 0){G=1/G;}
	B = log10((abs(format->channels[channel].blue_gain)+1)*10);
	if(format->channels[channel].blue_gain < 0){B=1/B;}
	//R = G = B = 1.0;
	if(format->orientation){
		if(!format->top_down_scan){
			n = format->pixels_per_line - n - 1;
		} 
		if(!format->left_right_scan){
			line = format->last_usable_line - line;
		}
		offset = n*input->pitch + line * 3;
	} else {
		if(!format->top_down_scan){
			line = format->last_usable_line - line;
		}
		if(!format->left_right_scan){
			n = format->pixels_per_line - n - 1;
		}
		offset = line*input->pitch + n * 3;
	}
	ov = (((double)pixels[offset]*B + (double)pixels[offset + 1]*G + (double)pixels[offset+2]*R)/765.0);
	if(ov > 1.0){ov = 1.0;}
	if(ov < 0.0){ov = 0.0;}
	//printf("%f ", ov);
	return ov;
	
}
/*Generates one line of NBTV signal*/
double generate_line(SDL_Surface *input, struct _input *p, struct NBTV_format *format, struct _output *opf, int channel, int line, int *op, double pixel_error, uint8_t *key, float *samples){
    double usps = 1000000.0 / (double)opf->sample_rate;
	double pixel_period = (double)round(10000000.0/(double)format->pixel_clock)/10.0;
    //printf("usps=%0.2f pc=%d pp=%0.2f\n", usps, format->pixel_clock, pixel_period);
	fflush(stdout);
    int n;
    int j;
    int missed_pixels = 0;
    double mpa = 0;
    double val;
	double pixel_error_integral=0;
	uint8_t *pixels=input->pixels;;
	    for(n = 0;n<format->pixels_per_line;n++){
		val =  get_converted_pixel(input, p, format, channel, line, n);
        for(j = 0;j < (((double)pixel_period+pixel_error+pixel_error_integral/3)/usps)-1;j++){
			if(key[*op]){
				//printf("Break because of key\n");
				goto exit;
			}
			if(line < format->first_usable_line || line > format->last_usable_line){
				samples[*op] = format->frame_blanking_value;
			} else {
				samples[*op] = val;
				if(missed_pixels){
					samples[*op] *= 0.5;
					samples[*op] += 0.5 * (mpa/missed_pixels);
					mpa = 0;
					missed_pixels = 0;
				}
			}
	//		printf("<%.1f> %.1f ",val,samples[*op]);
		//	if(format->channels[channel].){
		//		samples[*op] =- samples[*op];
		//	}
		//	samples[*op] *= (format->channels[channel].video_level/100.0);
	//		samples[*op] += (double)format->channels[channel].video_dc_off/127.0;
			if(samples[*op] > 1.0){samples[*op] = 1.0;}
			if(samples[*op] < -1.0){samples[*op] = -1.0;}
                (*op)++;
			}
            if(!j){
                missed_pixels++;
                mpa += val;
            }
            pixel_error += (pixel_period - ((double)j*usps));
			pixel_error_integral += pixel_error;
           // printf("pixel error = %0.1f pei = %0.1f j=%d op=%d\n", pixel_error, pixel_error_integral, j, *op);
    }
	exit:
	//printf("pixel period = %0.2f avg pixel erro = %0.2f\n", pixel_period, pixel_error_integral / (double)format->pixels_per_line);
    return pixel_error;
}

#define KEY_VIDEO 0
#define KEY_LINE_BLANK 1
#define KEY_FIRST_LINE_BLANK 2
#define KEY_LAST_LINE_BLANK 3
#define KEY_FRAME_BLANK 4	
/*Generates the key track. The key track defines where video data, blanking and sync go. All generated signals are generated
based on this track so multiple output tracks stay in sync*/
int generate_key_track(struct NBTV_format *format, struct _output *opf, uint8_t *key){
	int ps,ptr=0;
	int line;
	double usps = 1000000.0 / (double)opf->sample_rate;
	double error=0;
	int desired;
	double error_i=0;
	int last;
	int start;
	for(ptr = 0;ptr < ((double)format->frame_blanking_time/usps);ptr++){key[ptr] = KEY_FRAME_BLANK;}
	for(line = 0;line < format->lines;line++){
		start = ptr;
//		printf("error = %0.1f\n", error);
		for(ps = 0;ps < ((double)((double)format->line_blanking_time)/usps);ps++){
			if(line == 0){
				key[ptr + ps] = KEY_FIRST_LINE_BLANK;
			} else if(line == format->lines-1){
				key[ptr + ps] = KEY_LAST_LINE_BLANK;
			} else {
				key[ptr + ps] = KEY_LINE_BLANK;
			}
		}	
	last = ptr;
		ptr += ps;
		desired = (((1000000.0/(double)format->line_clock)+error+error_i/3)/usps)-ps;
		//printf("Desired = %d iE=%.1f", desired, error);
		for(ps = 0;ps < desired;ps++){
			key[ptr++] = KEY_VIDEO;
		}
		error += (1000000.0/(double)format->line_clock) - (((double)(ptr) - (double)start)*usps);
		error_i += error;
		//printf("Error = %.3f Error_i = %.1f ps=%d ptr = %d start = %d\n", error, error_i, ps, ptr, start);
	}
	return ptr;
}
	
int generate_frame(SDL_Surface *input, struct _input *p, struct NBTV_format *format, struct _output *opf, int channel, uint8_t *key, float *samples){
	double usps = 1000000.0 / (double)opf->sample_rate;
	struct NBTV_channel *ch = &(format->channels)[channel];
	int line;
	int op = 0;
	int dp=0;
	for(line = 0;line < format->lines;line++){
		while(key[op]){
			while(key[op] == KEY_FRAME_BLANK){
				if(!ch->fs_pulse || !ch->sync){
					samples[op] = ((double)format->frame_blanking_value/100.0) + ((double)ch->sync_dc_off/100.0);
				} else {
					samples[op] = ((double)ch->fs_level/100.0)*(double)(ch->pos_line_sync? 1.0: -1.0) + ((double)ch->sync_dc_off/100.0);
				}
				if(ch->invert_sync)	samples[op] =- samples[op];
				if(samples[op] > 1.0) samples[op] = 1.0;
				if(samples[op] < -1.0) samples[op] = -1.0;
				op++;
			}
			while(key[op] > 0 && key[op] < 4){
				if(!ch->line_sync || !ch->sync ||(key[op] == KEY_FIRST_LINE_BLANK && ch->fs_no_ls_f && ch->frame_sync) || (key[op] == KEY_LAST_LINE_BLANK && ch->fs_no_ls_l && ch->frame_sync)){
					samples[op] = ((double)format->line_blanking_value/100.0) + ((double)ch->sync_dc_off/100.0);
				} else {
					samples[op] = ((double)ch->ls_level/100.0)*(double)(ch->pos_line_sync? 1.0: -1.0) + ((double)ch->sync_dc_off/100.0);
					//printf("Sync sample: %.2f * %d + %.2f = %.2f\n", (double)ch->ls_level/100.0, ch->pos_line_sync? 1:-1, (double)ch->sync_dc_off/100.0, samples[op]);
				}
				if(ch->invert_sync) samples[op] =- samples[op];
				if(samples[op] > 1.0) samples[op] = 1.0;
				if(samples[op] < -1.0) samples[op] = -1.0;
				op++;
			}
			op++;
		}
				
		generate_line(input, p, format, opf, channel, line, &op, 0, key, samples);
//		if(op > 65000)break;
	}
	return op;
}
void generate_sync_track(struct _input *p, struct NBTV_format *format, struct _output *opf, int channel, int kl, uint8_t *key, float *samples){
	double usps = 1000000.0 / (double)opf->sample_rate;
	struct NBTV_channel *ch = &(format->channels)[channel];
	int line;
	int op = 0;
	int dp=0;
	int cnt;
	for(op=0;op<kl;op++){
		if(key[op] == KEY_FRAME_BLANK || (key[op] == KEY_FIRST_LINE_BLANK && format->frame_blanking_time == 0 && ch->fs_pulse)){
			for(cnt = 0;cnt < ch->fs_length/usps;cnt++){
				if(!ch->fs_pulse || !ch->sync || !ch->frame_sync){
					samples[op] = ((double)format->frame_blanking_value/100.0) + ((double)ch->sync_dc_off/100.0);
				} else {
					samples[op] = ((double)ch->fs_level/100.0)*(double)(ch->pos_line_sync? 1.0: -1.0) + ((double)ch->sync_dc_off/100.0);
				}
				if(ch->invert_sync)	samples[op] =- samples[op];
				if(samples[op] > 1.0) samples[op] = 1.0;
				if(samples[op] < -1.0) samples[op] = -1.0;
				op++;
			}
		} else if(key[op] > 0 && key[op] < 4){
			for(cnt = 0;cnt < ch->ls_length/usps;cnt++){
				if(!ch->line_sync || !ch->sync ||(key[op] == KEY_FIRST_LINE_BLANK && ch->fs_no_ls_f && ch->frame_sync) || (key[op] == KEY_LAST_LINE_BLANK && ch->fs_no_ls_l && ch->frame_sync)){
					samples[op] = ((double)format->line_blanking_value/100.0) + ((double)ch->sync_dc_off/100.0);
				} else {
					samples[op] = ((double)ch->ls_level/100.0)*(double)(ch->pos_line_sync? 1.0: -1.0) + ((double)ch->sync_dc_off/100.0);
					//printf("Sync sample: %.2f * %d + %.2f = %.2f\n", (double)ch->ls_level/100.0, ch->pos_line_sync? 1:-1, (double)ch->sync_dc_off/100.0, samples[op]);
				}
				if(ch->invert_sync) samples[op] =- samples[op];
				if(samples[op] > 1.0) samples[op] = 1.0;
				if(samples[op] < -1.0) samples[op] = -1.0;
				op++;
			}
		} else {
			samples[op] = 0;
		}
	}
}
extern int updated_buf;
int start_NBTV_gen(int seconds, struct _input *i, struct NBTV_format *f){
	if(!load_input_file(i)){
		error("Failed loading file %s\n", i->filename);
		return 0;
	}
	if(!seconds && !i->loop && i->length){
		target_timestamp = i->length + 81;
	} else {
		target_timestamp = seconds * 1000.0;
	}
	frame_changed = 1;
	format_changed = 1;
	input_timestamp = output_timestamp = 0;
	nbtv_gen = 1;
	start_output(&output);

}
int delay_time;
void stop_NBTV_gen(){
	nbtv_gen = 0;
	video_stop();
	stop_output();
	if(image){
		SDL_FreeSurface(image);
		image = NULL;
	}
	gui_update_output(&output);
}
void NBTV_gen_tick(){
	if(!nbtv_gen){
		SDL_Delay(50);
		return;
	} else {
		SDL_Delay(10);
	}
	if(format_changed){
		memset(key, 0, sizeof(uint8_t) * frame_len);
		frame_len=generate_key_track(nbtv_fmt, &output, key);
		debug("frame_len=%d\n", frame_len);
		fflush(stdout);
	}
	//printf("Input timestamp = %0.1f output timestamp = %0.1f d=%0.1f\n", input_timestamp, output_timestamp, input_timestamp - output_timestamp); 
	int ticks = SDL_GetTicks();
	if(ticks < delay_time && (output.format == OUTPUT_FORMAT_SC || input.multi_frame)){
		if(frame_changed && delay_time - ticks > 100){
		//	printf("Frame changed\n");
			update_input_preview(image, &input);
		}
		return;
	}
	delay_time = ticks + 950/nbtv_fmt->frame_rate;
	if(input_timestamp - output_timestamp > 150){
		delay_time -= (input_timestamp - output_timestamp);
		if(delay_time < 1){delay_time = 1;}
	}
	/*if(input_timestamp - output_timestamp > 150){
		delay_time *= 0.75;
	}*/
	if((frame_changed && (output_timestamp < input_timestamp || !input.multi_frame)) || (format_changed && !input.multi_frame)){
		window=frame_window(&input, nbtv_fmt);
		if(!window){return;}
		update_preview(image, window, &input);
		int k;
		for(k = 0;k < 4;k++){
			if(nbtv_fmt->channels[k].video){
				//memset(float_samples[k], 0, sizeof(float) * frame_len);
				generate_frame(window, &input, nbtv_fmt, &output, k, key, float_samples[k]);
				channel_zeroed[k] = 0;
			} else if(nbtv_fmt->channels[k].sync){
				generate_sync_track(&input, nbtv_fmt, &output, k, frame_len, key, float_samples[k]);
				channel_zeroed[k] = 0;
			} else if(!channel_zeroed[k]){
				memset(float_samples[k], 0, sizeof(float) * frame_len);
				channel_zeroed[k] = 1;
			}
		}

	//	SDL_LockAudio();
	//	SDL_UnlockAudio();

	/*	if(output_timestamp < input_timestamp){
			repeat_frame = 1;
		} else {
			repeat_frame = 0;
		}*/
		frame_changed = 0;
		format_changed = 0;
		repeat_frame = 0;
		nbtv_output(float_samples, key, frame_len, nbtv_fmt, &output);
		//frame_complete = 1;
		//SDL_Delay(900/nbtv_fmt->frame_rate);
		frame_count++;
	} else if(!input.multi_frame || input_timestamp - output_timestamp > 150){
		repeat_frame = 1;
		nbtv_output(float_samples, key, frame_len, nbtv_fmt, &output);
		/*if(output.format == OUTPUT_FORMAT_SC) SDL_Delay(900/nbtv_fmt->frame_rate);
		else SDL_Delay(10);*/
		frame_count++;
	}
	gui_update_time_produced();
	
/*	while(repeat_frame){
		nbtv_output(float_samples, key, frame_len, nbtv_fmt, &output);
	/*	if(output_timestamp < input_timestamp - (4000/nbtv_fmt->frame_rate)){
			repeat_frame = 0;
			output_timestamp = input_timestamp;
		}*/
	/*	if(output_timestamp >= input_timestamp){
			repeat_frame = 0;
		}
	}*/
	/*if(target_frame_count){
		if(!--target_frame_count){
			stop_NBTV_gen();
		}
	}*/
	if(target_timestamp && output_timestamp >= target_timestamp){
		stop_NBTV_gen();
	}
}
//void generate_frame
